/*********************************************************************
本ソースリストおよびソフトウェアは、ライセンスフリーです。(詳細は別記)
利用、編集、再配布等が自由に行えますが、著作権表示の改変は禁止します。

I2Cインタフェースの温度センサSTTS751と液晶を同じバス接続で動かす。

                               Copyright (c) 2014 Wataru KUNINO
                               http://www.geocities.jp/bokunimowakaru/
*********************************************************************/

#define	PORT_SCL	19					// I2C SCLポート番号(ANALOG 5)
#define	PORT_SDA	18					// I2C SDAポート番号(ANALOG 4)
#define	PORT_ERR	13					// エラー時のパルス出力用ポート
#define	DEBUG		1					// デバッグモード
#define	I2C_RAMDA	20					// I2C データシンボル長[us]

void i2c_debug(const char *s,byte priority){
	if(DEBUG||priority==5){
		Serial.print("[");
		Serial.print(millis()/1000,DEC);
		if(priority>3)	Serial.print("] ERROR:");
		else 			Serial.print("]      :");
		Serial.print(s);
		Serial.println();
	}
}

void i2c_error(const char *s){
	pinMode(PORT_ERR, OUTPUT);
	digitalWrite(PORT_ERR, HIGH);
	i2c_debug(s,5);
	digitalWrite(PORT_ERR, LOW);
}

void i2c_SCL(byte level){
	if( level ){
		pinMode(PORT_SCL, INPUT);
	}else{
		digitalWrite(PORT_SCL, LOW);
		pinMode(PORT_SCL, OUTPUT);
	}
	delayMicroseconds(I2C_RAMDA);
}

void i2c_SDA(byte level){
	if( level ){
		pinMode(PORT_SDA, INPUT);
	}else{
		digitalWrite(PORT_SDA, LOW);
		pinMode(PORT_SDA, OUTPUT);
	}
	delayMicroseconds(I2C_RAMDA);
}

byte i2c_tx(const byte in){
	byte i;
	for(i=0;i<8;i++){
		if( (in>>(7-i))&0x01 ){
				i2c_SDA(1);					// (SDA)	H Imp
		}else	i2c_SDA(0);					// (SDA)	L Out
		/*Clock*/
		i2c_SCL(1);							// (SCL)	H Imp
		i2c_SCL(0);							// (SCL)	L Out
	}
	/* ACK処理 */
	delayMicroseconds(I2C_RAMDA);
	i2c_SCL(1);								// (SCL)	H Imp
	i2c_SDA(1);								// (SDA)	H Imp
	for(i=50;i>0;i--){
		if( digitalRead(PORT_SDA) == 0 ) break;	// 速やかに確認
		delayMicroseconds(I2C_RAMDA);
	}
	if(i==0){
		i2c_error("I2C_TX / no ACK");
		return(0);
	}
	return(i);
}

byte i2c_init(void){
	byte i;

	for(i=50;i>0;i--){						// リトライ50回まで
		i2c_SDA(1);							// (SDA)	H Imp
		i2c_SCL(1);							// (SCL)	H Imp
		if( digitalRead(PORT_SCL)==1 &&
			digitalRead(PORT_SDA)==1  ) break;
		delay(1);
	}
	if(i==0) i2c_error("I2C_Init / Locked Lines");
	delayMicroseconds(I2C_RAMDA*8);
	return(i);
}

byte i2c_start(void){
	if(!i2c_init())return(0);				// SDA,SCL  H Out
	i2c_SDA(0);								// (SDA)	L Out
	delayMicroseconds(I2C_RAMDA);
	i2c_SCL(0);								// (SCL)	L Out
	return(1);
}

byte i2c_read(byte adr, byte *rx, byte len){
/*
入力：byte adr = I2Cアドレス
出力：byte *rx = 受信データ用ポインタ
入力：byte len = 受信長
戻り値：byte 受信結果長、０の時はエラー
			port=11		DIO11	pin  7		SCL
			port=12		DIO12	pin  4		SDA
・I2Cマスタ機能のみを実装（スレーブとしては動作しない）
・７ビットアドレスのみ対応
・１バイトの受信に５秒以上かかる（かなり遅いし、多量のXBee通信が発生する）
*/
	byte ret,i;
	
	if( !i2c_start() ) return(0);
	adr |= 0x01;							// RW=1 受信モード
	if( i2c_tx(adr)==0 ) return(0);			// アドレス設定
	
	/* スレーブ待機状態待ち */
	for(i=10;i>0;i--){
		delayMicroseconds(I2C_RAMDA);
		if( digitalRead(PORT_SDA)==0  ) break;
	}
	if(i==0){
		i2c_error("I2C_RX / no ACK");
		return(0);
	}
	for(i=10;i>0;i--){
		delayMicroseconds(I2C_RAMDA);
		if( digitalRead(PORT_SCL)==1  ) break;
	}
	if(i==0){
		i2c_error("I2C_RX / Clock Line Holded");
		return(0);
	}
	/* 受信データ */
	for(ret=0;ret<len;ret++){
		i2c_SCL(0);							// (SCL)	L Out
		i2c_SDA(1);							// (SDA)	H Imp
		rx[ret]=0x00;
		for(i=0;i<8;i++){
			i2c_SCL(1);						// (SCL)	H Imp
			rx[ret] |= (digitalRead(PORT_SDA))<<(7-i);		//data[22] b4=Port 12(SDA)
			i2c_SCL(0);						// (SCL)	L Out
		}
		// ACKを応答する
		i2c_SDA(0);							// (SDA)	L Out
		i2c_SCL(1);							// (SCL)	H Imp
		delayMicroseconds(I2C_RAMDA);
	}
	i2c_SDA(1);								// (SDA)	H Imp
	return(ret);
}

byte i2c_write(byte adr, byte *tx, byte len){
/*
入力：byte adr = I2Cアドレス ／SHT 温度測定時0x01	湿度測定時0x02
入力：byte *tx = 送信データ用ポインタ
入力：byte len = 送信データ長
*/
	byte i,ret;
	if(len){
		if( !i2c_start() ) return(0);
	}else return(0);
	adr &= 0xFE;							// RW=0 送信モード
	if( i2c_tx(adr)==0 ) return(0);			// アドレス設定

	/* データ送信 */
	for(ret=0;ret<len;ret++){
		i2c_SDA(0);							// (SDA)	L Out
		i2c_SCL(0);							// (SCL)	L Out
		i2c_tx(tx[ret]);
	}
	/* STOP */
	i2c_SDA(0);								// (SDA)	L Out
	i2c_SCL(0);								// (SCL)	L Out
	delayMicroseconds(I2C_RAMDA);
	i2c_SCL(1);								// (SCL)	H Imp
	i2c_SDA(1);								// (SDA)	H Imp
	return(ret);
}

void i2c_lcd_out(byte y,byte *lcd){
	byte data[2];
	byte i;
	data[0]=0x00;
	if(y==0) data[1]=0x80;
	else{
		data[1]=0xC0;
		y=1;
	}
	i2c_write(0x7C,data,2);
	for(i=0;i<8;i++){
		if(lcd[i]==0x00) break;
		data[0]=0x40;
		data[1]=lcd[i];
		i2c_write(0x7C,data,2);
	}
}

void utf_del_uni(char *s){
	byte i=0;
	byte j=0;
	while(s[i]!='\0'){
		if((byte)s[i]==0xEF){
			if((byte)s[i+1]==0xBE) s[i+2] += 0x40;
			i+=2;
		}
		s[j]=s[i];
		i++;
		j++;
	}
	s[j]='\0';
}

void i2c_lcd_print(char *s){
	byte i,j;
	byte lcd[9];
	
	utf_del_uni(s);
	lcd[8]='\0';
	for(j=0;j<2;j++){
		for(i=0;i<8;i++){
			lcd[i]=(byte)s[i+8*j];
			if(lcd[i]==0x00) for(;i<8;i++) lcd[i]=' ';
		}
		i2c_lcd_out(j,lcd);
	}
}

int i2c_temp(){
	/* 温度センサ STTS751 I2Cアドレス 0x72(Addr=33kΩ) */
	int temp;
	byte data;
	
	data=0x00;
	i2c_write(0x72,&data,1);		// レジスタ 0x00(温度上位桁)を指定
	i2c_read(0x72,&data,1);			// 読み出し
	temp = ((int)((signed char)data))*10;
	data=0x02;
	i2c_write(0x72,&data,1);		// レジスタ 0x00(温度上位桁)を指定
	i2c_read(0x72,&data,1);			// 読み出し
	temp += (((int)data/16)*10)/16;
	char s[32];
	sprintf(s,"Temp = %3d.%01d degC",temp/10,temp%10);
	i2c_debug(s,1);
	return(temp);
}

void setup(){
	byte data[2];
	Serial.begin(9600);
	Serial.println("Hello!");
	
	/* 液晶のセットアップ */
	i2c_init();			// I2Cインタフェースの初期化
	Serial.println("Init LCD");
	data[0]=0x00; data[1]=0x39; i2c_write(0x7C,data,2);	// IS=1
	data[0]=0x00; data[1]=0x11; i2c_write(0x7C,data,2);	// OSC
		/* OSC=0x14	0.21mA、OSC=0x11 0.18mA		*/
	data[0]=0x00; data[1]=0x70; i2c_write(0x7C,data,2);	// コントラスト	0
	data[0]=0x00; data[1]=0x56; i2c_write(0x7C,data,2);	// Power/Cont	6
	data[0]=0x00; data[1]=0x6C; i2c_write(0x7C,data,2);	// FollowerCtrl	C
	delay(200);
	data[0]=0x00; data[1]=0x38; i2c_write(0x7C,data,2);	// IS=0
	data[0]=0x00; data[1]=0x0C; i2c_write(0x7C,data,2);	// DisplayON	C
	data[0]=0x00; data[1]=0x01; i2c_write(0x7C,data,2);	// Clear Screen
	
	char s[]="ﾎﾞｸﾆﾓﾜｶﾙ XBee ZB";
	i2c_lcd_print(s);
	i2c_debug("Started",1);
}

void loop(){
	char s[32];
	
	delay(500);
	i2c_debug("Temperature --------",1);
	int temp=(int)i2c_temp();
	
	i2c_debug("LCD ----------------",1);
	sprintf(s,"ｵﾝﾄﾞ    %3d.%01dﾟC",temp/10,temp%10);
	utf_del_uni(s);
	i2c_lcd_print(s);
	i2c_debug("Done ---------------",1);
}

